import re

from czespressif.czespressif import CzPluginEspressif


def strip_ansi_codes(text: str) -> str:
    """Strip ANSI color codes from a string for snapshot comparison."""
    ansi_escape = re.compile(r'(?:\x1B[@-_][0-?]*[ -/]*[@-~])')
    return ansi_escape.sub('', text)


def test_example(config, snapshot):
    """Testing the output returned by 'cz example' and comparing it with snapshot"""
    cze_config = CzPluginEspressif(config)
    example = strip_ansi_codes(cze_config.example())
    assert example == snapshot


def test_schema(config, snapshot):
    """Testing the output returned by 'cz schema' and comparing it with snapshot"""
    cze_config = CzPluginEspressif(config)
    schema = cze_config.schema()
    assert schema == snapshot


def test_info(config, snapshot):
    """Testing the output returned by 'cz info' and comparing it with snapshot"""
    cze_config = CzPluginEspressif(config)
    info = strip_ansi_codes(cze_config.info())
    assert info == snapshot



from __future__ import annotations

from datetime import datetime
from typing import Literal

TYPES: list[dict] = [  # this is order in changelog
    {
        'type': 'BREAKING CHANGE',
        'description': 'Changes that are not backward-compatibles',
        'heading': 'Breaking changes',
        'emoji': '🚨',
        'bump': 'MAJOR',
        'regex': r'BREAKING[\-\ ]CHANGE',
        'question': False,  # Breaking changes have a dedicated question
        'changelog': True,
    },
    {
        'type': 'feat',
        'description': 'A new feature',
        'heading': 'New features',
        'emoji': '✨',
        'bump': 'MINOR',
        'changelog': True,
    },
    {
        'type': 'fix',
        'description': 'A bug fix',
        'heading': 'Bug fixes',
        'emoji': '🐛',
        'bump': 'PATCH',
        'changelog': True,
    },
    {
        'type': 'docs',
        'description': 'Documentation only change',
        'heading': 'Documentation',
        'emoji': '📖',
        'bump': 'PATCH',
        'changelog': True,
    },
    {
        'type': 'refactor',
        'description': 'A changeset neither fixing a bug nor adding a feature',
        'heading': 'Code refactoring',
        'emoji': '🔧',
        'bump': 'PATCH',
        'changelog': True,
    },
    {
        'type': 'remove',
        'description': 'Removing code or files',
        'heading': 'Removals',
        'emoji': '🗑️',
        'bump': 'PATCH',
        'changelog': True,
    },
    {
        'type': 'change',
        'description': 'A change made to the codebase.',
        'heading': 'Changes',
        'emoji': '🏗️',
        'bump': 'PATCH',
        'changelog': False,
    },
    {
        'type': 'ci',
        'description': 'Changes to CI configuration files and scripts',
        'heading': 'CI and project settings',
        'emoji': '⚙️',
        'bump': 'PATCH',
        'changelog': False,
    },
    {
        'type': 'test',
        'description': 'Adding missing or correcting existing tests',
        'heading': 'Testing',
        'emoji': '🚦',
        'bump': 'PATCH',
        'changelog': False,
    },
    {
        'type': 'revert',
        'description': 'Revert one or more commits',
        'heading': 'Reverted',
        'emoji': '🔙',
        'bump': 'PATCH',
        'changelog': False,
    },
]

CHANGELOG_TITLE: str = """
<a href="https://www.espressif.com">
    <img src="czespressif/templates/espressif-logo.svg" align="right" height="20" />
</a>

# CHANGELOG
"""

CHANGELOG_HEADER: str = """
> All notable changes to this project are documented in this file.
> This list is not exhaustive - only important changes, fixes, and new features in the code are reflected here.

<div align="center">
    <img alt="Static Badge" src="https://img.shields.io/badge/Keep%20a%20Changelog-v1.1.0-salmon?logo=keepachangelog&logoColor=black&labelColor=white&link=https%3A%2F%2Fkeepachangelog.com%2Fen%2F1.1.0%2F">
    <img alt="Static Badge" src="https://img.shields.io/badge/Conventional%20Commits-v1.0.0-pink?logo=conventionalcommits&logoColor=black&labelColor=white&link=https%3A%2F%2Fwww.conventionalcommits.org%2Fen%2Fv1.0.0%2F">
    <img alt="Static Badge" src="https://img.shields.io/badge/Semantic%20Versioning-v2.0.0-grey?logo=semanticrelease&logoColor=black&labelColor=white&link=https%3A%2F%2Fsemver.org%2Fspec%2Fv2.0.0.html">
</div>
<hr>
"""

CHANGELOG_FOOTER: str = f"""
<div align="center">
    <small>
        <b>
            <a href="https://www.github.com/espressif/cz-plugin-espressif">Commitizen Espressif plugin</a>
            ·
            <a href="https://www.github.com/espressif/standards">Espressif Standards</a>
        </b>
    <br>
        <sup><a href="https://www.espressif.com">Espressif Systems CO LTD. ({datetime.now().year})</a><sup>
    </small>
</div>
"""

INCREMENT = Literal['MAJOR', 'MINOR', 'PATCH']

# Currently BUMP_MESSAGE not possible to define default here - if not provided from pyproject.toml, it will be commitizen default, not ours :(
BUMP_MESSAGE: str = 'change(bump): release $current_version → $new_version [skip-ci]'

ESCAPE_MARKDOWN_SEQ: list[str] = [r'_{', r'}_']



from __future__ import annotations

import re
import sys

from dataclasses import dataclass
from dataclasses import field
from dataclasses import fields
from functools import cached_property
from functools import total_ordering
from typing import cast

from commitizen.config import read_cfg
from commitizen.defaults import Settings

from czespressif.defaults import CHANGELOG_FOOTER
from czespressif.defaults import CHANGELOG_HEADER
from czespressif.defaults import CHANGELOG_TITLE
from czespressif.defaults import INCREMENT
from czespressif.defaults import TYPES


@dataclass
@total_ordering
class CommitType:
    type: str  # Key used as type in the commit header.
    description: str  # A human-readable description of the type.
    heading: str | None = None  # The resulting heading in the changelog for this type.
    emoji: str | None = None  # An optional emoji representing the type.
    changelog: bool = True  # Whether this type should appear in the changelog or not.
    question: bool = True  # Whether this type should appear in the question choices.
    bump: INCREMENT | None = None  # Specifies the version bump category (e.g., 'MAJOR', 'MINOR', 'PATCH').
    regex: str | None = None  # An optional regular expression matching this type.

    def __str__(self) -> str:
        return self.type

    def __hash__(self) -> int:
        return hash(self.type)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, CommitType):
            return self._compare_key(other) == 0
        return NotImplemented

    def __lt__(self, other: object) -> bool:
        if isinstance(other, CommitType):
            return self._compare_key(other) < 0
        return NotImplemented

    def _compare_key(self, other: object) -> int:
        """Helper method for comparing with either CommitType or str."""
        if isinstance(other, CommitType):
            other_type = other.type.lower()
        elif isinstance(other, str):
            other_type = other.lower()
        else:
            return NotImplemented
        return (self.type.lower() > other_type) - (self.type.lower() < other_type)

    @classmethod
    def from_dict(cls, data: dict) -> CommitType:
        """Creates a CommitType instance from a dictionary."""
        valid_fields = {f.name for f in fields(cls)}
        filtered_data = {k: v for k, v in data.items() if k in valid_fields}
        return cls(**filtered_data)

    @classmethod
    def from_list(cls, lst: list[dict]) -> list[CommitType]:
        """Creates a list of CommitType instances from a list of dictionaries."""
        return [cls.from_dict(d) for d in lst]


class CzEspressifSettings(Settings):
    types: list[dict] | None
    extra_types: list[dict] | None
    changelog_unreleased: bool | None
    use_emoji: bool | None
    changelog_title: str | None
    changelog_header: str | None
    changelog_footer: str | None
    changelog_section_line: bool | None
    changelog_show_commits: bool | None
    changelog_show_authors: bool | None
    release_notes_footer: str | None


@dataclass
class CzEspressifConfig:
    settings: CzEspressifSettings = field(default_factory=lambda: read_cfg().settings)

    @property
    def types(self) -> list[CommitType]:
        return CommitType.from_list(self.settings.get('types', TYPES))

    @property
    def extra_types(self) -> list[CommitType]:
        return CommitType.from_list(self.settings.get('extra_types', []))

    @cached_property
    def known_types(self) -> list[CommitType]:
        return self.types + self.extra_types

    @property
    def incremental(self) -> bool:
        cli_version = any(re.match(r'^v?\d+\.\d+\.\d+$', arg) for arg in sys.argv)  # re: '(v)MAJOR.MINOR.PATCH'
        cli_version_range = any(re.match(r'^v?\d+\.\d+\.\d+\.\.v?\d+\.\d+\.\d+$', arg) for arg in sys.argv)  # re: '(v)MAJOR.MINOR.PATCH..(v)MAJOR.MINOR.PATCH'
        cli_incremental_flag = '--incremental' in sys.argv
        cli_bump = 'bump' in sys.argv

        is_incremental = any([cli_version, cli_version_range, cli_incremental_flag, cli_bump])
        return cast(bool, self.settings.get('incremental', is_incremental))

    @property
    def changelog_unreleased(self) -> bool:
        return cast(bool, self.settings.get('changelog_unreleased', True))

    @property
    def use_emoji(self) -> bool:
        return cast(bool, self.settings.get('use_emoji', True))

    @property
    def changelog_title(self) -> str:
        return cast(str, self.settings.get('changelog_title', CHANGELOG_TITLE))

    @property
    def changelog_header(self) -> str:
        return cast(str, self.settings.get('changelog_header', CHANGELOG_HEADER))

    @property
    def changelog_footer(self) -> str:
        return cast(str, self.settings.get('changelog_footer', CHANGELOG_FOOTER))

    @property
    def changelog_section_line(self) -> bool:
        return cast(bool, self.settings.get('changelog_section_line', True))

    @property
    def changelog_show_commits(self) -> bool:
        return cast(bool, self.settings.get('changelog_show_commits', True))

    @property
    def changelog_show_authors(self) -> bool:
        return cast(bool, self.settings.get('changelog_show_authors', True))

    @property
    def release_notes_footer(self) -> str:
        return cast(str, self.settings.get('release_notes_footer', ''))



from __future__ import annotations

import itertools
import re

from collections import OrderedDict
from collections.abc import Iterable
from typing import Any

from commitizen.cz.base import BaseCommitizen
from commitizen.cz.base import BaseConfig
from commitizen.defaults import Questions
from jinja2 import PackageLoader

from czespressif.config import CommitType
from czespressif.config import CzEspressifConfig
from czespressif.defaults import ESCAPE_MARKDOWN_SEQ
from czespressif.defaults import INCREMENT
from czespressif.example import build_example
from czespressif.example import build_info
from czespressif.questions import get_questions


class CzPluginEspressif(BaseCommitizen):  # pylint: disable=abstract-method
    template_loader = PackageLoader('czespressif', 'templates')

    def __init__(self, config: BaseConfig):
        super().__init__(config)
        self.cze_config = CzEspressifConfig(config.settings)

    @property
    def known_types(self) -> dict[str, CommitType]:
        return dict(itertools.chain((t.type, t) for t in self.cze_config.known_types))

    @property
    def re_known_types(self) -> str:
        return '|'.join(t.regex or k for k, t in self.known_types.items())

    @property
    def changelog_pattern(self) -> str:
        re_known_types = '|'.join(k for k, t in self.known_types.items())
        return rf'\A({re_known_types})(\(.+\))?(!)?'

    @property
    def change_type_map(self) -> dict[str, str]:
        return dict(
            itertools.chain(
                ((t.type, f'{t.emoji + " " if self.cze_config.use_emoji and t.emoji else ""}{t.heading}') for t in self.cze_config.known_types if t.heading),
            )
        )

    @property
    def change_type_order(self) -> list[str]:
        return [
            f'{commit_type.emoji} {commit_type.heading}' if self.cze_config.use_emoji and commit_type.emoji else commit_type.heading
            for commit_type in self.cze_config.known_types
            if commit_type.changelog and commit_type.heading
        ]

    @property
    def bump_pattern(self) -> str:
        """Regex to extract information from commit (subject and body)"""
        re_types = '|'.join(t.regex or k for k, t in self.known_types.items() if t.bump)
        return rf'^((({re_types})(\(.+\))?(!)?)|\w+!):'

    @property
    def bump_map(self) -> dict[str, INCREMENT]:
        """
        Mapping the extracted information to a SemVer increment type (MAJOR, MINOR, PATCH)
        """
        return OrderedDict(
            (
                (r'^.+!$', 'MAJOR'),
                *((rf'^{t.regex or k}', t.bump) for k, t in self.known_types.items() if t.bump),
            )
        )

    @property
    def bump_map_major_version_zero(self) -> dict[str, INCREMENT]:
        return OrderedDict((pattern, increment.replace('MAJOR', 'MINOR')) for pattern, increment in self.bump_map.items())  # type: ignore

    @property
    def commit_parser(self) -> str:
        return (
            rf'(?P<change_type>{self.re_known_types})'
            r'(?:\((?P<scope>[^()\r\n]*)\))?'  # Optional scope
            r'(?P<breaking>!)?:\s'
            r'(?P<message>.*)'  # Required message
            r'(?:\n{2,}(?P<body>.*))?'  # Optional body
            r'(?:\n{2,}(?P<footer>.*))?'  # Optional footer
        )

    @property
    def template_extras(self) -> dict[str, Any]:
        return {'config': self.cze_config, 'settings': self.cze_config.settings}

    def questions(self) -> Questions:
        return get_questions(self.cze_config)

    def message(self, answers: dict[str, Any]) -> str:
        prefix = answers['type']
        scope = answers['scope']
        subject = answers['subject']
        body = answers['body']
        is_breaking_change = answers['is_breaking_change']
        breaking_change = answers.get('breaking_change')
        footer = answers.get('footer', '')
        extra = ''

        if scope:
            scope = f'({scope})'

        if is_breaking_change and not breaking_change:
            scope += '!'

        if body:
            body = f'\n\n{body}'

        if is_breaking_change and breaking_change:
            footer = f'BREAKING CHANGE: {breaking_change}\n{footer}'

        if footer:
            footer = f'\n\n{footer}'

        message = f'{prefix}{scope}: {subject}{extra}{body}{footer}'

        return message

    def wrap_problematic_parts(self, message: str) -> str:
        """Escape problematic underscores near curly braces without altering other underscores."""
        # Dynamically build regex pattern to handle all known issues
        for issue in ESCAPE_MARKDOWN_SEQ:
            # For each known issue, replace underscores with escaped underscores
            message = re.sub(issue, lambda match: match.group(0).replace('_', r'\_'), message)
        return message

    def changelog_message_builder_hook(self, parsed_message: dict[str, Any], commit: Any) -> dict[str, Any] | Iterable[dict[str, Any]] | None:  # pylint: disable=unused-argument
        # Remap breaking changes type
        if parsed_message.get('breaking'):
            parsed_message['change_type'] = 'BREAKING CHANGE'

        # Get the full list of types to include from pyproject.toml
        included_in_changelog = self.cze_config.settings.get('types_in_changelog', [])

        # If the list is empty (i.e., not set), include all default types
        if not included_in_changelog:
            commit_type = self.known_types.get(parsed_message['change_type'])
            if commit_type and commit_type.changelog:
                # Apply the escaping to the parsed message
                parsed_message['message'] = self.wrap_problematic_parts(parsed_message['message'])
                return parsed_message
            return None

        # Filter out all commit types not explicitly included in `types_in_changelog`
        if parsed_message['change_type'] not in included_in_changelog:
            return None

        # Apply the escaping to the parsed message
        parsed_message['message'] = self.wrap_problematic_parts(parsed_message['message'])

        return parsed_message

    def changelog_hook(self, full: str, partial: str | None) -> str:
        """Process resulting changelog to keep 1 empty line at the end of the file."""
        changelog = partial or full
        return changelog.rstrip() + '\n'

    def example(self) -> str:
        examples = build_example(types=self.cze_config.types, extra_types=self.cze_config.extra_types)
        return examples

    def schema(self) -> str:
        return '<type>(<scope>): <subject>\n' '<BLANK LINE>\n' '<body>\n' '<BLANK LINE>\n' '(BREAKING CHANGE: <breaking changes>)*\n' '(<footers>)*'

    def schema_pattern(self) -> str:
        pattern = (
            r'(?s)'  # To explicitly make . match new line
            rf'({self.re_known_types})'  # type
            r'(\(\S+\))?!?:'  # scope
            r'( [^\n\r]+)'  # subject
            r'((\n\n.*)|(\s*))?$'
        )
        return pattern

    def info(self) -> str:
        return build_info()


discover_this = CzPluginEspressif  # pylint: disable=invalid-name



from __future__ import annotations

from dataclasses import dataclass
from random import getrandbits

import pytest

from commitizen.config import BaseConfig
from commitizen.git import GitCommit
from syrupy.extensions.single_file import SingleFileSnapshotExtension
from syrupy.extensions.single_file import WriteMode

from czespressif.config import CzEspressifConfig
from czespressif.config import CzEspressifSettings


class MarkdownSnapshotExtension(SingleFileSnapshotExtension):
    _file_extension = 'md'
    _write_mode = WriteMode.TEXT


@pytest.fixture
def snapshot(snapshot):
    return snapshot.use_extension(MarkdownSnapshotExtension)


def randbytes(size: int) -> bytes:
    return bytearray(getrandbits(8) for _ in range(size))


@dataclass
class Factory:
    config: CzEspressifConfig

    def parsed_message(self, **kwargs) -> tuple[dict, GitCommit]:
        parsed = {'type': 'change', 'scope': None, 'message': 'I am a message', **kwargs}
        prefix = parsed['type']
        msg = [f'{prefix}: {parsed["message"]}']
        body = parsed.get('body')
        if body is not None:
            msg.extend(('', body))
        footer = parsed.get('footer')
        if footer is not None:
            msg.extend(('', footer))
        return parsed, self.commit('\n'.join(msg))

    def commit(self, title: str, **kwargs) -> GitCommit:
        return GitCommit(rev=str(randbytes(8)), title=title, **kwargs)


@pytest.fixture
def settings(request) -> CzEspressifSettings:
    settings = CzEspressifSettings()
    for marker in reversed(list(request.node.iter_markers('settings'))):
        settings.update(marker.kwargs)
    return settings


@pytest.fixture
def config(settings):
    config = BaseConfig()
    config.settings.update({'name': 'czespressif'})
    config.settings.update(settings)
    return config


@pytest.fixture
def cze_config(settings) -> CzEspressifConfig:
    return CzEspressifConfig(settings)


@pytest.fixture
def factory(cze_config: CzEspressifConfig) -> Factory:
    return Factory(cze_config)



from __future__ import annotations

from pathlib import Path

import pytest

from commitizen import git
from commitizen.changelog import Metadata
from commitizen.commands.changelog import Changelog
from commitizen.exceptions import DryRunExit
from pytest_mock import MockerFixture
from syrupy.extensions.single_file import SingleFileSnapshotExtension
from syrupy.extensions.single_file import WriteMode

FIXTURES = Path(__file__).parent / 'fixtures/changelogs'


class MarkdownSnapshotExtension(SingleFileSnapshotExtension):
    _file_extension = 'md'
    _write_mode = WriteMode.TEXT


@pytest.fixture
def snapshot(snapshot):
    return snapshot.use_extension(MarkdownSnapshotExtension)


COMMITS_DATA = [
    {
        'rev': '22c4820bd927ad017649bcf6e8d03e30ddc946e9',
        'title': 'change(bump): [skip-ci] release v1.2.0 → v1.3.0',
        'body': '',
        'author': 'Jane Hunter',
        'author_email': 'jane.hunter@espressif.com',
    },
    {
        'rev': 'b5e7a2d4c9f547d8b5c7f2a9e7c5d4a8e2d6b7c4',
        'title': 'feat(api): rollback changes to JWT token handling',
        'body': '',
        'author': 'Alice Johnson',
        'author_email': 'alice.johnson@espressif.com',
    },
    {
        'rev': 'd9b2c7e8f4b946a7f5c7a5b8d7f3e4c2b5a6f9d3',
        'title': 'fix(logging): switch from console logs to structured JSON logging',
        'body': '',
        'author': 'Eve Martin',
        'author_email': 'eve.martin@espressif.com',
    },
    {
        'rev': 'f2e7d1b7c4b942a5a5d3c2f9a7f5b4d9e5c3b7e4',
        'title': 'test(api): write unit tests for new password reset feature',
        'body': '',
        'author': 'Eve Martin',
        'author_email': 'eve.martin@espressif.com',
    },
    {
        'rev': 'b7a4c5b2d5e147c8b5f7a2e5d4c9e3b2a7f4d9e5',
        'title': 'fix: correct typos in error messages displayed on UI',
        'body': '',
        'author': 'Bob Smith',
        'author_email': 'bob.smith@espressif.com',
    },
    {
        'rev': 'c5d3e7a2f2b945c7a5e7d2b9f5c4e3b8d7f9e2a4',
        'title': 'feat(ui): improve loading animations for dashboard page',
        'body': '',
        'author': 'Diana White',
        'author_email': 'diana.white@espressif.com',
    },
    {
        'rev': 'e2d9a7f5b5e24c5f9b7d4f2c5a7e9b4c3d5e6f2a',
        'title': 'ci: add automated deployment script for production',
        'body': '',
        'author': 'Charlie Green',
        'author_email': 'charlie.green@espressif.com',
    },
    {
        'rev': 'd25d90e96d5f6bdeeca1f340cce9f7b87f7c8c0e',
        'title': 'change(bump): [skip-ci] release v1.2.0.rc0 → v1.2.0',
        'body': '',
        'author': 'Jane Hunter',
        'author_email': 'jane.hunter@espressif.com',
    },
    {
        'rev': 'd7e2f9b5c4b24b5a9a7b3d2f5c4e7f9c2b8e5d1f',
        'title': 'ci: configure linter to enforce new coding standards',
        'body': '',
        'author': 'Diana White',
        'author_email': 'diana.white@espressif.com',
    },
    {
        'rev': 'f5c7b1e4e4b947d9a7b5f2c5e4f9d7a8b5c3e7d9',
        'title': 'change(bump): [skip-ci] release v1.1.0 → v1.2.0',
        'body': '',
        'author': 'Alice Johnson',
        'author_email': 'alice.johnson@espressif.com',
    },
    {
        'rev': 'e7c1d8b6e2b549a8a5c6b3e2d9a7f5c4e2d7b6a9',
        'title': 'refactor: optimize database queries for user data retrieval',
        'body': '',
        'author': 'Charlie Green',
        'author_email': 'charlie.green@espressif.com',
    },
    {
        'rev': '05d02b594c146c89cf2c5d5c1e773da93fdcd1a3',
        'title': 'change(bump): [skip-ci] release v1.1.0 → v1.2.0.rc0',
        'body': '',
        'author': 'Jane Hunter',
        'author_email': 'jane.hunter@espressif.com',
    },
    {
        'rev': 'f8d7e1c27f0e4935bb5b5e8f5c4e3b0d4e8b6c4e',
        'title': 'docs: update API documentation for user endpoint',
        'body': '',
        'author': 'Eve Martin',
        'author_email': 'eve.martin@espressif.com',
    },
    {
        'rev': 'c3e7b3b9d7e645e9b7a4f5b7e2c8e9c4f8b5d4e9',
        'title': 'fix(frontend): correct layout issues in user profile page',
        'body': '',
        'author': 'Bob Smith',
        'author_email': 'bob.smith@espressif.com',
    },
    {
        'rev': '0b17d79a2baf456a72e04f8986469d8a69a845b1',
        'title': 'change(bump): [skip-ci] release v1.0.0 → v1.1.0',
        'body': '',
        'author': 'Ethan Mitchell',
        'author_email': 'ethan.mitchell@espressif.com',
    },
    {
        'rev': '1b7d1e7f2d8444a9bf7f5c547e8b9a3f66e4c8d1',
        'title': 'fix(api): update response codes for failed logins',
        'body': '',
        'author': 'Bob Smith',
        'author_email': 'bob.smith@espressif.com',
    },
    {
        'rev': '3d5b9c71e82b44b9b5c6a8e23f9e456fb7d5a4c1',
        'title': 'test: add integration tests for payment gateway module',
        'body': '',
        'author': 'Diana White',
        'author_email': 'diana.white@espressif.com',
    },
    {
        'rev': 'd4a8fbc34eae47c9871b5cb83e7db791f4c2c7b0',
        'title': 'revert: remove deprecated functions from auth service',
        'body': '',
        'author': 'Alice Johnson',
        'author_email': 'alice.johnson@espressif.com',
    },
    {
        'rev': '8f0b3b9f1e2a4b84a5b5c789e2c6a8d4d3b5c6f0',
        'title': 'remove: delete old caching logic from request handlers',
        'body': '',
        'author': 'Charlie Green',
        'author_email': 'charlie.green@espressif.com',
    },
    {
        'rev': 'de8dfc7d65f1e9ed8fd264dbe62380a520e9853f',
        'title': 'change(bump): [skip-ci] release v0.2.1 → v1.0.0',
        'body': '',
        'author': 'Jane Hunter',
        'author_email': 'jane.hunter@espressif.com',
    },
    {
        'rev': '7c6d91fae8bcd5239a3f8c4973e7e843f93c1be0',
        'title': 'docs(frontend): Update README for new build system integration',
        'body': '',
        'author': 'Diana White',
        'author_email': 'diana.white@espressif.com',
    },
    {
        'rev': 'c8e7a2d4b0fb55ea9b6f55b637f31af8e2f9d935',
        'title': 'ci(pipeline): add test for new database migrations',
        'body': '',
        'author': 'Charlie Green',
        'author_email': 'charlie.green@espressif.com',
    },
    {
        'rev': 'e2d5bca6e1f047a9bfcd7b72c39d3147d3b2d1ea',
        'title': 'change: refactor code to use async/await syntax',
        'body': '',
        'author': 'Eve Martin',
        'author_email': 'eve.martin@espressif.com',
    },
    {
        'rev': 'f3c34c19173b142b0fc0140a50ab5c60317e522a',
        'title': 'change(bump): [skip-ci] release v0.2.0 → v0.2.1',
        'body': '',
        'author': 'Victor Moore',
        'author_email': 'victor.moore@espressif.com',
    },
    {
        'rev': 'b2c7e8a4fc76d68baf9a5b8a7de5d3b7d4cbb745',
        'title': 'feat(api): implement new authentication mechanism with JWT tokens',
        'body': '',
        'author': 'Alice Johnson',
        'author_email': 'alice.johnson@espressif.com',
    },
    {
        'rev': '4b1d62bfc7a942e8b123b17f8cba07d5ec4de1e6',
        'title': 'fix: resolve null pointer exception in user session management',
        'body': '',
        'author': 'Bob Smith',
        'author_email': 'bob.smith@espressif.com',
    },
    {
        'rev': '18e5faebc94d1aecc3fd7c8cfb63d5197800041f',
        'title': 'change(bump): [skip-ci] release v0.1.0 → v0.2.0',
        'body': '',
        'author': 'Nina Shaw',
        'author_email': 'nina.shaw@espressif.com',
    },
    {
        'rev': 'b2c7e8a4fc76d68baf9a5b8a7de5d3b7d4cbb745',
        'title': 'feat(api): implement new authentication mechanism with JWT tokens',
        'body': '',
        'author': 'Alice Johnson',
        'author_email': 'alice.johnson@espressif.com',
    },
    {
        'rev': '4b1d62bfc7a942e8b123b17f8cba07d5ec4de1e6',
        'title': 'fix: resolve null pointer exception in user session management',
        'body': '',
        'author': 'Bob Smith',
        'author_email': 'bob.smith@espressif.com',
    },
    {
        'rev': '7d1f8c5bd1b56a65e30e6bb7f5dd1c2c1b0e5d4c',
        'title': 'ci(init): initialize project with empty commit, initial tag',
        'body': '',
        'author': 'Tomas Sebestik',
        'author_email': 'tomas.sebestik@espressif.com',
    },
]


TAGS = [
    ('v1.3.0', '22c4820bd927ad017649bcf6e8d03e30ddc946e9', '2024-08-30'),
    ('v1.2.0', 'd25d90e96d5f6bdeeca1f340cce9f7b87f7c8c0e', '2024-07-15'),
    ('v1.2.0.rc0', '05d02b594c146c89cf2c5d5c1e773da93fdcd1a3', '2024-06-10'),
    ('v1.1.0', '0b17d79a2baf456a72e04f8986469d8a69a845b1', '2024-05-23'),
    ('v1.0.0', 'de8dfc7d65f1e9ed8fd264dbe62380a520e9853f', '2024-04-12'),
    ('v0.2.1', 'f3c34c19173b142b0fc0140a50ab5c60317e522a', '2024-03-29'),
    ('v0.2.0', '18e5faebc94d1aecc3fd7c8cfb63d5197800041f', '2024-03-02'),
    ('v0.1.0', '7d1f8c5bd1b56a65e30e6bb7f5dd1c2c1b0e5d4c', '2024-02-15'),
]


@pytest.fixture
def gitcommits() -> list[git.GitCommit]:
    return [git.GitCommit(**commit) for commit in COMMITS_DATA]


@pytest.fixture
def tags() -> list[git.GitTag]:
    tags = [git.GitTag(*tag) for tag in TAGS]
    return tags


@pytest.fixture
def render_changelog(config, gitcommits, tags, capsys, mocker: MockerFixture):
    """
    Generate a changelog using the same flow as the cz changelog command.
    """

    def fixture(unreleased: bool = True, **kwargs) -> str:
        incremental = kwargs.get('incremental', config.settings.get('incremental', False))

        # Mock Git commits based on whether it's incremental or full
        mocker.patch('commitizen.git.get_commits', return_value=gitcommits[:4] if incremental else gitcommits)
        mocker.patch('commitizen.git.get_tags', return_value=tags)

        kwargs['dry_run'] = True
        kwargs['incremental'] = incremental
        kwargs['unreleased_version'] = unreleased

        cmd = Changelog(config, kwargs)
        mocker.patch.object(cmd.changelog_format, 'get_metadata').return_value = Metadata(latest_version='1.3.0')

        capsys.readouterr()

        try:
            cmd()
        except DryRunExit:
            pass

        return capsys.readouterr().out

    return fixture


### Czespressif plugin
### -------------------------------------------------


def test_changelog_czespressif_full(render_changelog, snapshot):
    """
    DEFAULT SETTINGS - Full changelog
    Emulates command:   'cz changelog' ('cz ch')
    With pyproject.toml:
                        [tool.commitizen]
                            name                       = "czespressif"
                            changelog_merge_prerelease = true
                            tag_format                 = "v$version"
    """
    assert render_changelog() == snapshot


@pytest.mark.settings(incremental=True)
def test_changelog_czespressif_incremental(render_changelog, snapshot):
    """
    DEFAULT SETTINGS - Incremental changelog
    Emulates command:   'cz changelog v1.3.0' ('cz ch v1.3.0')
                        'cz changelog v1.2.0..v1.3.0' ('cz ch v1.2.0..1.3.0')
                        'cz bump'
    With pyproject.toml:
                        [tool.commitizen]
                            name                       = "czespressif"
                            changelog_merge_prerelease = true
                            tag_format                 = "v$version"
    """
    assert render_changelog() == snapshot


@pytest.mark.settings(use_emoji=False)
def test_changelog_czespressif_full_no_emoji(render_changelog, snapshot):
    """
    WITHOUT EMOJI - Full changelog
    Emulates command:   'cz changelog' ('cz ch')
    With pyproject.toml:
                        [tool.commitizen]
                            name                       = "czespressif"
                            changelog_merge_prerelease = true
                            tag_format                 = "v$version"
                            use_emoji                  = false
    """
    assert render_changelog() == snapshot


@pytest.mark.settings(use_emoji=False)
@pytest.mark.settings(incremental=True)
def test_changelog_czespressif_incremental_no_emoji(render_changelog, snapshot):
    """
    WITHOUT EMOJI - Incremental changelog
    Emulates command:   'cz changelog v1.3.0' ('cz ch v1.3.0')
                        'cz changelog v1.2.0..v1.3.0' ('cz ch v1.2.0..1.3.0')
                        'cz bump'
    With pyproject.toml:
                        [tool.commitizen]
                            name                       = "czespressif"
                            changelog_merge_prerelease = true
                            tag_format                 = "v$version"
                            use_emoji                  = false
    """
    assert render_changelog() == snapshot


### WITHOUT 'czespressif' plugin
### -------------------------------------------------


@pytest.mark.settings(name='cz_conventional_commits')  # This is the default value for Commitizen
def test_changelog_cz_default_full(render_changelog, snapshot):
    """
    DEFAULT SETTINGS COMMITIZEN WITHOUT czespressif PLUGIN - Full changelog
    Emulates command:   'cz changelog' ('cz ch')
    With pyproject.toml:
                        [tool.commitizen]
                            changelog_merge_prerelease = true
                            tag_format                 = "v$version"
    """
    assert render_changelog() == snapshot


@pytest.mark.settings(name='cz_conventional_commits')  # This is the default value for Commitizen
def test_changelog_cz_default_incremental(render_changelog, snapshot):
    """
    DEFAULT SETTINGS COMMITIZEN WITHOUT czespressif PLUGIN - Incremental changelog
    Emulates command:   'cz changelog v1.3.0' ('cz ch v1.3.0')
                        'cz changelog v1.2.0..v1.3.0' ('cz ch v1.2.0..1.3.0')
                        'cz bump'
    With pyproject.toml:
                        [tool.commitizen]
                            changelog_merge_prerelease = true
                            tag_format                 = "v$version"
    """
    assert render_changelog(incremental=True) == snapshot



from __future__ import annotations

import pytest

from commitizen.changelog import Metadata
from commitizen.commands.changelog import Changelog
from commitizen.config import BaseConfig
from commitizen.exceptions import DryRunExit
from syrupy.extensions.single_file import SingleFileSnapshotExtension
from syrupy.extensions.single_file import WriteMode


# Using Syrupy to compare markdown snapshots
class MarkdownSnapshotExtension(SingleFileSnapshotExtension):
    _file_extension = 'md'
    _write_mode = WriteMode.TEXT


@pytest.fixture
def snapshot(snapshot):
    return snapshot.use_extension(MarkdownSnapshotExtension)


def render_changelog(config, gitcommits, tags, capsys, mocker, incremental: bool) -> str:
    mocker.patch('commitizen.git.get_commits', return_value=gitcommits[:4] if incremental else gitcommits)
    mocker.patch('commitizen.git.get_tags', return_value=tags)

    kwargs = {'dry_run': True, 'incremental': incremental, 'unreleased_version': True}
    cmd = Changelog(config, kwargs)

    mocker.patch.object(cmd.changelog_format, 'get_metadata').return_value = Metadata(latest_version='1.3.0')
    capsys.readouterr()

    try:
        cmd()
    except DryRunExit:
        pass

    return capsys.readouterr().out


@pytest.mark.parametrize('incremental', [False, True])
def test_changelog(snapshot, incremental, mocker, gitcommits, tags, capsys):
    # Create config for changelog
    config = BaseConfig()
    config.settings.update({'name': 'czespressif'})

    # Render the changelog using the provided fixture
    changelog = render_changelog(config, gitcommits, tags, capsys, mocker, incremental)

    # Compare the output with the markdown snapshot (Syrupy automatically handles snapshot naming)
    assert changelog == snapshot



from __future__ import annotations

from typing import Any

from commitizen.cz import exceptions

from czespressif.config import CzEspressifConfig


def required_validator(answer: str, msg: str = '') -> str:
    if not answer:
        raise exceptions.AnswerRequiredError(msg)
    return answer


def multiple_line_breaker(answer: str, sep: str = '|') -> str:
    return '\n'.join(line.strip() for line in answer.split(sep) if line)


def parse_scope(text: str) -> str:
    if not text:
        return ''
    scope = text.strip().split()
    if len(scope) == 1:
        return scope[0]
    return '-'.join(scope)


def parse_subject(text: str) -> str:
    if isinstance(text, str):
        text = text.strip('.').strip()
    return required_validator(text, msg='Subject is required.')


def get_questions(cze_config: CzEspressifConfig) -> list[dict[str, Any]]:
    """Questions regarding the commit message."""

    type_padding = 10  # Define the padding for the type part
    description_column = 25  # Define where the description should start
    extra_space_emoji = ['change', 'ci', 'remove']  # Types that need extra space behind the emoji

    return [
        {
            'type': 'list',
            'name': 'type',
            'message': 'Select the type of change you are committing',
            'choices': [
                {
                    'value': t.type,
                    # Add an extra space behind the emoji for specific types
                    'name': f'{t.type:<{type_padding}}'
                    + ' ' * (description_column - len(f'{t.type:<{type_padding}}'))
                    + f'{t.emoji}{"  " if t.type in extra_space_emoji else " "} {t.description}',
                }
                for t in cze_config.known_types
                if t.question
            ],
        },
        {
            'type': 'input',
            'name': 'scope',
            'message': 'Scope. Define Could be anything specifying place of the commit change (wifi, bt, core):\n',
            'filter': parse_scope,
        },
        {
            'type': 'input',
            'name': 'subject',
            'filter': parse_subject,
            'message': 'Subject. Concise description of the changes. Imperative, lower case and no period:\n',
        },
        {
            'type': 'input',
            'name': 'body',
            'message': 'Provide additional contextual information - commit message BODY: (press [enter] to skip)\n',
            'filter': multiple_line_breaker,
        },
        {
            'type': 'confirm',
            'name': 'is_breaking_change',
            'message': 'Is this a BREAKING CHANGE?\n',
            'default': False,
        },
        {
            'when': lambda x: x['is_breaking_change'],
            'type': 'input',
            'name': 'breaking_change',
            'message': 'Breaking changes. Details the breakage:\n',
            'filter': multiple_line_breaker,
        },
        {
            'type': 'input',
            'name': 'footer',
            'message': 'Footer. Information about Breaking Changes and reference issues that this commit impacts: (press [enter] to skip)\n',
            'filter': multiple_line_breaker,
        },
    ]



from __future__ import annotations

from czespressif.config import CommitType

# ANSI escape codes for colors and bold text
YELLOW = '\033[93m'
GREEN = '\033[92m'
RED = '\033[91m'
BOLD = '\033[1m'
LIGHT_BLUE = '\033[96m'
PURPLE = '\033[35m'
BLUE = '\033[94m'
RESET = '\033[0m'


def build_example(types: list[CommitType], extra_types: list[CommitType]) -> str:
    """
    'cz example' command output.
    Generate an example guide for commit message schema and conventions.
    # Exclude 'BREAKING CHANGE' from the known types list
    """

    # Exclude 'BREAKING CHANGE' from the known types list
    filtered_types = [t for t in types if t.type != 'BREAKING CHANGE']

    # Format the known types section with two spaces of indentation
    types_formatted = '\n'.join(f'  {GREEN}{t.type}{RESET}: {t.description}' for t in filtered_types)
    extra_types_formatted = '\n'.join(f'  {BLUE}{t.type}{RESET}: {t.description}' for t in extra_types)
    if extra_types_formatted:
        extra_types_formatted += f'  {BLUE}<--- non standard type added by extra_types{RESET}'

    # Example guide with colors and bold text
    example_guide = f"""
---

{BOLD}{YELLOW}Commit message schema:{RESET}

    {YELLOW}<type>(<scope>): <subject>{RESET}
        <... empty line ...>
    {YELLOW}<body>{RESET}
        <... empty line ...>
    {YELLOW}(BREAKING CHANGE: <breaking changes>)*{RESET}
    {YELLOW}(<footers>)*{RESET}

---

{BOLD}{GREEN}Commit types in this project:{RESET}

{types_formatted}
{extra_types_formatted}

---

{BOLD}{LIGHT_BLUE}Short commit messages (one line):{RESET}

    {LIGHT_BLUE}fix{RESET}: fix something (present or imperative, no period)
    {LIGHT_BLUE}fix(package){RESET}: fix something
    {LIGHT_BLUE}feat(name){RESET}: add a new feature
    {LIGHT_BLUE}refactor(package){RED}!{RESET}: this refactoring {RED}breaks the thing{RESET}

---

{BOLD}{BLUE}Full commit messages with body and footer:{RESET}

    {BLUE}feat(something){RESET}: add something new (present/imperative, no period)

    Here optionally comes the details - this is commit message body.
    It can be multiline, hyphens, asterisks, are okay.

    {RED}BREAKING CHANGE: It breaks something!{RESET}
    {BLUE}Closes https://github.com/espressif/<project>/issues/<issue_number>{RESET}

---

{BOLD}{PURPLE}More info:{RESET}

- https://www.conventionalcommits.org/en/v1.0.0/#specification
- https://github.com/{PURPLE}espressif/conventional-precommit-linter{RESET}
- https://github.com/{PURPLE}espressif/standards{RESET}
    """

    return example_guide


def build_info() -> str:
    """
    'cz info' command output.
    Returns information about a custom Commitizen plugin for Espressif Systems.
    """
    return (
        'This is a custom Commitizen plugin for Espressif Systems. \n'
        'It is mainly used to generate consistent and beautiful changelog with minimum settings from target project. \n'
        'Can be used for creating commit messages, that passes Espressif standards. \n'
        'Customization from target project is possible, check this plugin GitHub repo'
    )
